@@ -31,102 +31,104 @@ logger = logging.getLogger(__name__)
 
 
 class BaseBenchmark(ABC):
-    '''Abstract base class for evaluating models.
+    '''Abstract base class for finetuning and evaluating models.
 
     Args:
         name (str):
             The name of the dataset.
         task (str):
             The type of task to be benchmarked.
-        num_labels (int or None, optional):
-            The number of labels in the dataset. Defaults to None.
-        label2id (dict or None, optional):
-            A dictionary that converts labels to their indices. This will only
-            be used if the pretrained model does not already have one. Defaults
-            to None.
+        metric_names (dict):
+            A dictionary with the variable names of the metrics used in the
+            dataset as keys, and a more human readable name of them as values.
+        id2label (list or None, optional):
+            A list of all the labels, which is used to convert indices to their
+            labels. This will only be used if the pretrained model does not
+            already have one. Defaults to None.
+        label_synonyms (list of lists of str or None, optional):
+            A list of synonyms for each label. Every entry in `label_synonyms`
+            is a list of synonyms, where one of the synonyms is contained in
+            `id2label`. If None then no synonyms will be used. Defaults to
+            None.
         evaluate_train (bool, optional):
             Whether the models should be evaluated on the training scores.
             Defaults to False.
         cache_dir (str, optional):
             Where the downloaded models will be stored. Defaults to
             '.benchmark_models'.
-        learning_rate (float, optional):
-            What learning rate to use when finetuning the models. Defaults to
-            2e-5.
-        epochs (int, optional):
-            The number of epochs to finetune for. Defaults to 5.
-        warmup_steps (int, optional):
-            The number of training steps in which the learning rate will be
-            warmed up, meaning starting from nearly 0 and progressing up to
-            `learning_rate` after `warmup_steps` many steps. Defaults to 50.
-        batch_size (int, optional):
-            The batch size used while finetuning. Must be a multiple of 2, and
-            at most 32. Defaults to 32.
+        two_labels (bool, optional):
+            Whether two labels should be predicted in the dataset.  If this is
+            True then `split_point` has to be set. Defaults to False.
+        split_point (int or None, optional):
+            When there are two labels to be predicted, this is the index such
+            that `id2label[:split_point]` contains the labels for the first
+            label, and `id2label[split_point]` contains the labels for the
+            second label. Only relevant if `two_labels` is True. Defaults to
+            None.
         verbose (bool, optional):
             Whether to print additional output during evaluation. Defaults to
             False.
 
     Parameters:
+        name (str): The name of the dataset.
         task (str): The type of task to be benchmarked.
-        num_labels (int or None): The number of labels in the dataset.
-        label2id (dict or None): A dictionary converting labels to indices.
+        metric_names (dict): The names of the metrics.
         id2label (dict or None): A dictionary converting indices to labels.
+        label2id (dict or None): A dictionary converting labels to indices.
+        num_labels (int or None): The number of labels in the dataset.
+        label_synonyms (list of lists of str): Synonyms of the dataset labels.
+        evaluate_train (bool): Whether the training set should be evaluated.
         cache_dir (str): Directory where models are cached.
-        learning_rate (float): Learning rate used while finetuning.
-        epochs (int): The number of epochs to finetune for.
-        warmup_steps (int): Number of steps used to warm up the learning rate.
-        batch_size (int): The batch size used while finetuning.
+        two_labels (bool): Whether two labels should be predicted.
+        split_point (int or None): Splitting point of `id2label` into labels.
         verbose (bool): Whether to print additional output.
-
-    Raises:
-        TypeError:
-            If `batch_size` is not among 1, 2, 4, 8, 16 or 32.
     '''
     def __init__(self,
                  name: str,
                  task: str,
                  metric_names: Dict[str, str],
-                 num_labels: Optional[int] = None,
                  id2label: Optional[List[str]] = None,
                  label_synonyms: Optional[List[List[str]]] = None,
                  evaluate_train: bool = False,
                  cache_dir: str = '.benchmark_models',
-                 learning_rate: float = 2e-5,
-                 epochs: int = 5,
-                 warmup_steps: int = 50,
-                 batch_size: int = 32,
-                 multilabel: bool = False,
+                 two_labels: bool = False,
                  split_point: Optional[int] = None,
                  verbose: bool = False):
-        if batch_size not in [1, 2, 4, 8, 16, 32]:
-            raise TypeError('The batch size must be either 1, 2, 4, 8, '
-                            '16 or 32.')
-        self.batch_size = batch_size
-        self.gradient_accumulation = 32 // batch_size
+
         self.name = name
-        self.metric_names = metric_names
-        self.evaluate_train = evaluate_train
         self.task = task
-        self.cache_dir = cache_dir
-        self.learning_rate = learning_rate
-        self.epochs = epochs
-        self.warmup_steps = warmup_steps
-        self.num_labels = num_labels
+        self.metric_names = metric_names
         self.id2label = id2label
-        self.multilabel = multilabel
+        self.label_synonyms = label_synonyms
+        self.evaluate_train = evaluate_train
+        self.cache_dir = cache_dir
+        self.two_labels = two_labels
         self.split_point = split_point
         self.verbose = verbose
-        self.label_synonyms = label_synonyms
-        if self.id2label is not None:
-            if label_synonyms is not None:
-                self.label2id = {label: id for id, lbl in enumerate(id2label)
-                                 for label_syns in label_synonyms
-                                 for label in label_syns
-                                 if lbl in label_syns}
-            else:
-                self.label2id = {lbl: id for id, lbl in enumerate(id2label)}
+
+        if id2label is not None:
+
+            # Store the number of labels
+            self.num_labels = len(id2label)
+
+            # Set default value of label synonyms, if None was given
+            if label_synonyms is None:
+                self.label_synonyms = [[label] for label in self.id2label]
+
+            # Define the label2id conversion dictionary
+            self.label2id = {label: id for id, lbl in enumerate(id2label)
+                             for label_syns in self.label_synonyms
+                             for label in label_syns
+                             if lbl in label_syns}
+
+        # If the id2label conversion list was not given, then set the number of
+        # labels to zero and set the label2id conversion dict to None as well
         else:
+            self.num_labels = None
             self.label2id = None
+
+        # If verbose is set to True then enable transformers output, which is
+        # done by setting it to warning (the default)
         if verbose:
             tf_logging.set_verbosity_warning()
 
@@ -309,28 +311,35 @@ class BaseBenchmark(ABC):
                 # have not been trained on (it will just always get those
                 # labels wrong)
                 else:
+                    # Collect the dataset labels and model labels in the
+                    # `model_id2label` conversion list
                     for label in self.id2label:
-                        if label not in model_id2label:
+                        syns = [syn for lst in self.label_synonyms
+                                for syn in lst
+                                if label in lst]
+                        if all([syn not in model_id2label for syn in syns]):
                             model_id2label.append(label)
 
-                    if self.label_synonyms is None:
-                        synonyms = [[label] for label in model_id2label]
-                    else:
-                        synonyms = self.label_synonyms
+                    # Get the synonyms of all the labels, new ones included
+                    new_synonyms = self.label_synonyms
+                    flat_old_synonyms = [syn for lst in self.label_synonyms
+                                         for syn in lst]
+                    new_synonyms += [[label] for label in model_id2label
+                                     if label not in flat_old_synonyms]
 
                     # Add all the synonyms of the labels into the label2id
                     # conversion dictionary
                     model_label2id = {label: id
                                       for id, lbl in enumerate(model_id2label)
-                                      for label_syns in synonyms
+                                      for label_syns in new_synonyms
                                       for label in label_syns
                                       if lbl in label_syns}
 
                     # Get the old id2label conversion
                     if not isinstance(model.config.id2label, list):
-                        old_model_id2label = dict(model.config.id2label)
+                        old_id2label = dict(model.config.id2label)
                     else:
-                        old_model_id2label = model.config.id2label
+                        old_id2label = model.config.id2label
 
                     # This changes the classification layer in the finetuned
                     # model to be consistent with all the labels in the
@@ -343,12 +352,23 @@ class BaseBenchmark(ABC):
                     #       needs to be rewritten when we add other types of
                     #       tasks.
                     # NOTE: Only works for pytorch models at the moment
-                    if (len(model_id2label) > len(old_model_id2label)
+                    if (len(model_id2label) > len(old_id2label)
                             and framework == 'pytorch'):
 
                         # Count the number of new labels to add to the model
                         num_new_labels = (len(model_id2label) -
-                                          len(old_model_id2label))
+                                          len(old_id2label))
+
+                        # If *all* the new labels are new and aren't even
+                        # synonyms of the model's labels, then raise an
+                        # exception
+                        if num_new_labels == self.num_labels:
+                            if len(set(flat_old_synonyms)
+                                    .intersection(old_id2label)) == 0:
+                                msg = ('The model has not been trained on '
+                                       'any of the labels in the dataset, or '
+                                       'synonyms thereof.')
+                                raise InvalidBenchmark(msg)
 
                         # Load the weights from the model's current
                         # classification layer. This handles both the token
@@ -467,13 +487,16 @@ class BaseBenchmark(ABC):
 
     @abstractmethod
     def _compute_metrics(self,
-                         predictions_and_labels: tuple) -> Dict[str, float]:
+                         predictions_and_labels: tuple,
+                         id2label: Optional[list] = None) -> Dict[str, float]:
         '''Compute the metrics needed for evaluation.
 
         Args:
             predictions_and_labels (pair of arrays):
                 The first array contains the probability predictions and the
                 second array contains the true labels.
+            id2label (list or None, optional):
+                Conversion of indices to labels. Defaults to None.
 
         Returns:
             dict:
@@ -595,7 +618,6 @@ class BaseBenchmark(ABC):
 
     def __call__(self,
                  model_id: str,
-                 num_finetunings: int = 10,
                  progress_bar: bool = True
                  ) -> Dict[str, List[Dict[str, float]]]:
         '''Finetune and evaluate a model.
@@ -604,10 +626,6 @@ class BaseBenchmark(ABC):
             model_id (str):
                 The full HuggingFace Hub path to the pretrained transformer
                 model.
-            num_finetunings (int, optional):
-                The number of times to finetune. These results will be used to
-                calculate confidence intervals of the means of the metrics.
-                Defaults to 10.
             progress_bar (bool, optional):
                 Whether to show a progress bar or not. Defaults to True.
 
@@ -636,8 +654,7 @@ class BaseBenchmark(ABC):
 
         # Get bootstrap sample indices
         if finetune or self.evaluate_train:
-            train_bidxs = rng.integers(0, len(train),
-                                       size=(num_finetunings - 1, len(train)))
+            train_bidxs = rng.integers(0, len(train), size=(9, len(train)))
         test_bidxs = rng.integers(0, len(test), size=(9, len(test)))
 
         if framework in ['pytorch', 'tensorflow', 'jax']:
@@ -662,7 +679,7 @@ class BaseBenchmark(ABC):
             trains = [train]
             if finetune or self.evaluate_train:
                 trains += [Dataset.from_dict(train[train_bidxs[idx]])
-                           for idx in range(num_finetunings - 1)]
+                           for idx in range(9)]
             tests = [test]
             tests += [Dataset.from_dict(test[test_bidxs[idx]])
                       for idx in range(test_bidxs.shape[0])]
@@ -670,9 +687,9 @@ class BaseBenchmark(ABC):
             # Set up progress bar
             if finetune:
                 if progress_bar:
-                    itr = tqdm(range(num_finetunings))
+                    itr = tqdm(range(10))
                 else:
-                    itr = range(num_finetunings)
+                    itr = range(10)
             else:
                 itr = [0]
 
@@ -692,11 +709,12 @@ class BaseBenchmark(ABC):
                 save_strategy='epoch',
                 report_to='none',
                 save_total_limit=1,
-                per_device_train_batch_size=self.batch_size,
-                learning_rate=self.learning_rate,
-                num_train_epochs=self.epochs,
-                warmup_steps=self.warmup_steps,
-                gradient_accumulation_steps=self.gradient_accumulation,
+                per_device_train_batch_size=32,
+                per_device_eval_batch_size=32,
+                learning_rate=2e-5,
+                num_train_epochs=1000,
+                warmup_steps=(len(train) // 4),
+                gradient_accumulation_steps=1,
                 load_best_model_at_end=True
             )
 
@@ -719,7 +737,8 @@ class BaseBenchmark(ABC):
                         )
 
                         # Initialise early stopping callback
-                        params = dict(early_stopping_patience=2)
+                        patience = 2 + 250 // len(train)
+                        params = dict(early_stopping_patience=patience)
                         early_stopping = EarlyStoppingCallback(**params)
 
                         # Initialise Trainer
@@ -732,7 +751,7 @@ class BaseBenchmark(ABC):
                                             data_collator=data_collator,
                                             compute_metrics=compute_metrics,
                                             callbacks=[early_stopping])
-                        if self.multilabel:
+                        if self.two_labels:
                             trainer_args['split_point'] = self.split_point
                             trainer = TwolabelTrainer(**trainer_args)
                         else:
@@ -750,7 +769,7 @@ class BaseBenchmark(ABC):
                         # Log training metrics and save the state
                         if self.evaluate_train:
                             train_metrics = trainer.evaluate(
-                                tests[idx],
+                                trains[idx],
                                 metric_key_prefix='train'
                             )
                             metrics['train'].append(train_metrics)
@@ -765,7 +784,7 @@ class BaseBenchmark(ABC):
 
                         break
 
-                    except RuntimeError as e:
+                    except (RuntimeError, ValueError) as e:
                         if not str(e).startswith('CUDA out of memory'):
                             raise InvalidBenchmark(str(e))
                         bs = training_args.per_device_train_batch_size
@@ -832,7 +851,7 @@ class BaseBenchmark(ABC):
                 train = self._preprocess_data(train, framework=framework)
                 trains = [train]
                 trains += [Dataset.from_dict(train[train_bidxs[idx]])
-                           for idx in range(num_finetunings - 1)]
+                           for idx in range(9)]
 
                 # Get the train predictions
                 all_train_metrics = list()
